from .queues import ReadQueue, WriteQueue, Queue, SimpleQueue, QueueKV
from .pipelines import Pipeline, Wrapped, Workflow, PipelineQueues, make_queues, flatten_queues

__all__ = [
  'ReadQueue', 'WriteQueue', 'Queue', 'SimpleQueue', 'QueueKV',
  'Pipeline', 'Wrapped', 'Workflow',
  'PipelineQueues', 'make_queues', 'flatten_queues'
]


from typing_extensions import AsyncIterable, Generic, TypeVar
import asyncio
import os
from datetime import timedelta
from haskellian import iter as I, Either, Left, IsLeft, Right
from kv.api import KV, InexistentItem as KVInexistentItem, DBError
from pipeteer.queues import Queue, QueueError, ReadError, InexistentItem

A = TypeVar('A')
B = TypeVar('B')

class QueueKV(Queue[A], Generic[A]):
  """Poll-based queue backed by a `KV` object"""
  def __init__(
    self, kv: KV[A],
    poll_interval: timedelta = timedelta(seconds=5),
  ):
    self._kv = kv
    self.poll_interval = poll_interval

  @classmethod
  def of(cls, conn_str: str, type: type[A] | None = None) -> 'QueueKV[A]':
    """Create a `Queue` with a `KV` from a connection string. Supports:
    - `file://<path>`: `FilesystemKV`
    - `sql+<protocol>://<conn_str>;Table=<table>`: `SQLKV`
    - `azure+blob://<conn_str>`: `BlobKV`
    - `azure+blob+container://<conn_str>;Container=<container_name>`: `BlobContainerKV`
    """
    from kv.api import KV
    return QueueKV(KV.of(conn_str, type))
  
  @classmethod
  def sqlite(cls, Type: type[B], path: str, table: str = 'queue') -> 'QueueKV[B]':
    try:
      from kv.sqlite import SQLiteKV
    except ImportError:
      raise ImportError('Install `kv-sqlite-sync` to run in local SQLite')
    return QueueKV(SQLiteKV.validated(Type, path, table))
  
  async def push(self, key: str, value: A):
    return (await self._kv.insert(key, value)).mapl(QueueError)
  
  async def _read(self, id: str | None = None) -> Either[ReadError, tuple[str, A]]:
    if id is None:
      key = I.head((await self._kv.keys()).unsafe())
      if key is None:
        await asyncio.sleep(self.poll_interval.total_seconds())
        return await self._read(id)
    else:
      key = id
    
    try:
      value = (await self._kv.read(key)).unsafe()
      return Right((key, value))
    except IsLeft as e:
      match e.value:
        case KVInexistentItem():
          return Left(InexistentItem(key))
        case err:
          return Left(QueueError(err))
        
  async def pop(self, id: str) -> Either[QueueError | InexistentItem, None]:
    match await self._kv.delete(id):
      case Left(KVInexistentItem()):
        return Left(InexistentItem(id))
      case Left(err):
        return Left(QueueError(err))
      case _:
        return Right(None)
    
  def _items(self) -> AsyncIterable[Either[QueueError, tuple[str, A]]]:
    return self._kv.items().map(lambda e: e.mapl(QueueError))
  


from .specs import Pipeline, Wrapped, Workflow
from .queues import make_queues, flatten_queues, PipelineQueues

__all__ = [
  'Pipeline', 'Wrapped', 'Workflow',
  'PipelineQueues', 'make_queues', 'flatten_queues',
]


from typing_extensions import TypeVar, TypedDict, Generic, Callable, Mapping, Sequence, Iterable, NotRequired
from pipeteer.queues import ReadQueue, WriteQueue, Queue, ops
from pipeteer.pipelines import Pipeline, Wrapped, Workflow

A = TypeVar('A')
B = TypeVar('B')
T = TypeVar('T')

class PipelineQueues(TypedDict, Generic[A, B]):
  """Possibly nested pipeline queues"""
  Qin: ReadQueue[A]
  Qout: WriteQueue[B]
  internal: Mapping[str, 'PipelineQueues'] | 'PipelineQueues'

def make_queues(
  Qin: ReadQueue[A], Qout: WriteQueue[B], pipeline: Pipeline[A, B], *,
  make_queue: Callable[[Sequence[str], type[T]], Queue[T]], prefix: tuple[str, ...] = ()
) -> PipelineQueues[A, B]:
  """Connect all queues of a `pipeline` into a nested tree"""
  match pipeline:
    case Workflow() as workflow:
      Qins = {
        task: (
          Qin if task == workflow.input_task else
          make_queue(prefix + (task,), workflow.pipelines[task].Tin)
        )
        for task in workflow.pipelines
      }
      inner_Qout = ops.prejoin(Qout, [(Qin, workflow.pipelines[id].Tin) for id, Qin in Qins.items()]) # type: ignore
      internal = {
        task: make_queues(Qins[task], inner_Qout, workflow.pipelines[task], make_queue=make_queue, prefix=prefix + (task,))
        for task in workflow.pipelines
      }
      return PipelineQueues(Qin=Qin, Qout=Qout, internal=internal)

    case Wrapped() as wrapped:
      inner_Qin = ops.immutable(Qin).map(wrapped.pre)
      inner_Qout = ops.premerge(Qin, Qout, wrapped.post)
      internal = make_queues(inner_Qin, inner_Qout, wrapped.pipeline, make_queue=make_queue, prefix=prefix + ('wrapped',))
      return PipelineQueues(Qin=Qin, Qout=Qout, internal=internal)
    
    case Pipeline():
      return PipelineQueues(Qin=Qin, Qout=Qout) # type: ignore
    
def flatten_queues(queues: PipelineQueues, *, prefix: tuple[str, ...] = ()) -> Iterable[tuple[Sequence[str], ReadQueue]]:
  """Flatten a nested tree of queues into a sequence of `(path, queue)` pairs"""
  for id, q in queues.items():
    if isinstance(q, dict):
      yield from flatten_queues(q, prefix + (id,)) # type: ignore
    elif isinstance(q, ReadQueue):
      yield prefix + (id,), q


from typing_extensions import TypeVar, Generic, Callable, Mapping
from dataclasses import dataclass

A = TypeVar('A')
B = TypeVar('B')
S1 = TypeVar('S1')
S2 = TypeVar('S2')

@dataclass
class Pipeline(Generic[A, B]):
  Tin: type[A]
  Tout: type[B] | None = None

@dataclass
class Wrapped(Pipeline[S1, S2], Generic[S1, S2, A, B]):
  def __init__(
    self, Tin: type[S1], pipeline: Pipeline[A, B],
    pre: Callable[[S1], A], post: Callable[[S1, B], S2]
  ):
    self.Tin = Tin
    self.pipeline = pipeline
    self.pre = pre
    self.post = post

@dataclass
class Workflow(Pipeline[A, B], Generic[A, B]):
  def __init__(self, Tin: type[A], Tout: type[B] | None = None, *, pipelines: Mapping[str, Pipeline]):
    
    input_tasks = [id for id, pipe in pipelines.items() if issubclass(pipe.Tin, Tin)]
    if len(input_tasks) > 1:
      raise ValueError(f'Workflow has multiple tasks with input type {Tin}: {input_tasks}')
    if len(input_tasks) == 0:
      raise ValueError(f'Workflow has no tasks with input type {Tin}')
    self.input_task = input_tasks[0]
    
    self.Tin = Tin
    self.Tout = Tout
    self.pipelines = pipelines

  def __repr__(self):
    out = f'Workflow({self.Tin.__name__},\n'
    for id, pipe in self.pipelines.items():
      out += f'  {id}: {repr(pipe)},\n'
    return out + ')'